import asyncio
from asyncio.exceptions import CancelledError
import fnmatch
import json
import os
import sys
from datetime import datetime, timedelta, timezone
from pathlib import Path
from shutil import rmtree
from typing import Any, Callable, List, Tuple, Dict

import boto3
import click
import urwid
from urwid.main_loop import AsyncioEventLoop

from seesaw import storage
from seesaw.config import LOG_DIR, LOG_FILE, SETTINGS, USER_DIR, logger
from seesaw.log_factory import (
    AsyncLogStreamPoller,
    groups_from_region,
    ordered_log_query,
)
from seesaw.transforms import (
    abbreviate_group_name,
    format_event,
    humanize_size,
    localize_timestamp,
    relative_utc_time,
    unmask_group_name,
)

log = logger(__file__)


@click.group(invoke_without_command=True)
@click.option("--region", default=None)
@click.option("-m", "--backfill-minutes", default=30)
@click.pass_context
def cli(ctx, region, backfill_minutes):
    log.info("launched", context=vars(ctx), region=region)
    if not backfill_minutes > 0:
        print("-m/--backfill-minutes must be a positive number")
        sys.exit(1)

    if ctx.invoked_subcommand is None:
        account_id = boto3.client("sts", region_name=region).get_caller_identity()[
            "Account"
        ]
        storage.remove_old(region, account_id, timedelta(days=7))
        try:
            tui_loop(region, account_id, backfill=backfill_minutes)
        except KeyboardInterrupt:
            sys.exit(3)


@cli.command("files")
@click.option("--clear", default=False, is_flag=True)
@click.option("-v", "--verbose", default=False, is_flag=True)
def files(clear, verbose):
    print("Log file", LOG_FILE)
    print("Config file", str(USER_DIR / "config.json"))
    print("Databases", str(USER_DIR / f"*.sqlite"))

    total_size = 0

    for top in (LOG_DIR, USER_DIR):
        for cwd, _, files in os.walk(top):
            cwd = Path(cwd)
            for f in files:
                size = os.stat(cwd / f).st_size
                if verbose:
                    print(humanize_size(size), (cwd / f))
                total_size += size
    print("Total space:", humanize_size(total_size))

    if clear:
        if LOG_DIR.exists():
            print("Deleting seesaw logs")
            rmtree(LOG_DIR)
        if USER_DIR.exists():
            print("Deleting seesaw user data")
            rmtree(USER_DIR)


@cli.command("tail")
@click.option("--region", default=None)
@click.option("-a", "--all-groups", is_flag=True, help="Tail all existing log groups")
@click.option(
    "-l",
    "--list-groups",
    is_flag=True,
    metavar="list_groups",
    help="List groups that would be tailed, then exit. Useful for testing glob patterns.",
)
@click.option(
    "--long-dates/--short-dates",
    is_flag=True,
    help="Show full dates, or shorten to just the time.",
)
@click.argument("LOG_GROUPS", nargs=-1)
def tail(region, all_groups, long_dates, list_groups, log_groups):
    """Pass any number LOG_GROUP names or fnmatch patterns to select log groups.

    - * - matches everything

    - ? - matches any single character

    - [seq] - matches any character in seq

    - [!seq] - matches any character not in seq

    Example: view all lambda logs

    $ seesaw tail '/aws/lambda*'

    Example: view all logs from MyApp log groups, including any Lambda, EC2, or Step Functions groups

    $ seesaw tail '*MyApp*'

    Example: view all API Gateway Execution logs

    $ seesaw tail 'API-Gateway-Exec*'"""
    if region is None:
        region = SETTINGS["presets"]["default"]["region"]
    if not log_groups:
        log_groups = SETTINGS["presets"]["default"]["log_groups"]
    if any(c in "".join(log_groups) for c in list("*[]?!")) or all_groups:
        # if the log groups have any globbing characters in them, run a glob match
        group_list = [g["name"] for g in groups_from_region(region)]
        if all_groups:
            log_groups = group_list
        else:
            matched_groups = [
                group
                for group in group_list
                if any(fnmatch.fnmatchcase(group, pattern) for pattern in log_groups)
            ]

            log_groups = matched_groups
        log.info("Polling all available log groups", groups=log_groups)
    if list_groups:
        for l in log_groups:
            print(l)
        return

    try:
        asyncio.run(tail_loop(region, log_groups, long_dates=long_dates))
    except KeyboardInterrupt:
        pass


async def task_monitor():
    prior = {}
    while True:
        await asyncio.sleep(3)
        states = {t.get_name(): t.done() for t in asyncio.all_tasks()}
        if states != prior:
            log.info(
                "active coroutines",
                tasks=sorted(t.get_name() for t in asyncio.all_tasks()),
                states=states,
                source="task_monitor",
            )
            prior = states


def create_log_group_aio_pollers(region, log_groups, account_id):
    tasks = {t.get_name(): t for t in asyncio.all_tasks()}
    for task_name, coroutine in tasks.items():
        if not task_name.startswith("GroupPoller"):
            continue
        polled_group = task_name.split(" ", 1)[-1]
        if polled_group not in log_groups and not coroutine.done():
            # then we can cancel the aio task
            coroutine.cancel("No longer required")
            log.info(
                "Cancelled polling for group",
                group=polled_group,
                task_name=task_name,
                stack=coroutine.get_stack(),
                cancel_status=coroutine.cancelled(),
            )

    for group in log_groups:
        if any(group in t for t in tasks):
            # there is already a poller for this log group, skip it
            continue
        group_poller = AsyncLogStreamPoller(
            region,
            account_id=account_id,
            start_time=datetime.now(tz=timezone.utc) - timedelta(minutes=5),
        )
        asyncio.create_task(group_poller.slurp_logs(group), name=f"GroupPoller {group}")


async def tail_loop(region, log_groups, long_dates):
    asyncio.create_task(
        task_monitor(),
        name="task_monitor",
    )
    asyncio.current_task().set_name("tail_loop")
    local_account = boto3.client("sts").get_caller_identity()["Account"]
    print("AWS account:", local_account, "region:", region, file=sys.stderr)
    print("groups:", ", ".join(log_groups), file=sys.stderr)
    log.debug(
        "Commencing tail of", region=region, groups=log_groups, account=local_account
    )
    create_log_group_aio_pollers(region, log_groups, local_account)

    async for event in ordered_log_query(
        region, local_account, start_from=timedelta(minutes=5), log_groups=log_groups
    ):
        print(" ".join(format_event(event, verbose_dates=long_dates)).strip())


async def refresh_pane(new_message: Callable, **kwargs):
    sent = 0
    log.debug("refreshing visible logs")
    try:
        async for event in ordered_log_query(**kwargs):
            sent += 1
            log.debug("Pushing event to pane", evt=event.gid, total_this_setting=sent)
            new_message(event)
    except CancelledError:
        pass
    except:
        log.exception("Refresh error")


class LogReadingPane:
    def __init__(self, region: str, account: str, backfill: int) -> None:
        self.backfill = backfill
        self.region = region
        self.account = account
        self.messages = urwid.SimpleFocusListWalker([])
        self.viewer = urwid.ListBox(self.messages)
        self.verbose_dates = False
        self.help_active = False
        self.backfill_tasks: Dict[str, Any] = {}

        self.poller = None

        self.menu = GroupSelect(
            region=region, account=account, groups_changed=self.create_aio_poll_task
        )
        self.top = urwid.Columns([(70, self.menu), self.viewer], dividechars=2)

    def create_aio_poll_task(self, groups):
        if self.poller:
            log.info("Killing poller", t=type(self.poller))
            self.poller.cancel("Replacing with new poller")
            log.debug(
                "Cancelled one", t={t.get_name(): t.done() for t in asyncio.all_tasks()}
            )
            self.messages.clear()
        for g in groups:
            if g in self.backfill_tasks:
                # we already are running backfill
                continue
            # start backfill tasks, breaking up into 15 minute chunks
            self.backfill_tasks[g] = []
            for i in range(0, self.backfill, 15):
                log.info("Backfill", start=i, end=i + 15)
                group_poller = AsyncLogStreamPoller(
                    self.region,
                    account_id=self.account,
                    poll_delay_seconds=1,
                    start_time=relative_utc_time(minutes=-i),
                )
                self.backfill_tasks[g].append(
                    asyncio.create_task(
                        group_poller.slurp_logs(
                            g, end_time=relative_utc_time(minutes=-(i + 15))
                        ),
                        name=f"BackfillPoller {i}-{i+15} {g}",
                    )
                )

        self.poller = asyncio.get_running_loop().create_task(
            refresh_pane(
                new_message=self.new_message,
                start_from=timedelta(hours=2),
                region=self.region,
                account_id=self.account,
                log_groups=groups,
            ),
            name=f"pane-refresher-{datetime.now().strftime('%H:%M:%S.%f')}",
        )

    def new_message(self, event):
        if len(self.messages) > 500:
            self.messages.pop(0)
        self.messages.append(urwid.Text(self.format_event(event)))
        self.viewer.focus_position = len(self.messages) - 1

    def toggle_date(self):
        self.verbose_dates = not self.verbose_dates

    def help_popup(self):
        self.help_active = not self.help_active

    def timestamp(self, ts):
        if self.verbose_dates:
            return (
                "timestamp",
                localize_timestamp(ts).strftime("%Y-%m-%d %H:%M:%S"),
            )
        return (
            "timestamp",
            localize_timestamp(ts).strftime("%H:%M:%S"),
        )

    def format_event(self, event: storage.Event):
        try:

            return [
                self.timestamp(event.timestamp),
                " ",
                abbreviate_group_name(event.log_group),
                " ",
                json.dumps(json.loads(event.message), indent=2, sort_keys=True),
            ]
        except:
            return [
                self.timestamp(event.timestamp),
                " ",
                abbreviate_group_name(event.log_group),
                " ",
                event.message,
            ]


class GroupSelect(urwid.WidgetWrap):
    def __init__(self, account: str, region: str, groups_changed: Callable):
        self.region = region
        self.account = account
        self.selected_groups = set()
        self.groups_changed = groups_changed
        urwid.WidgetWrap.__init__(self, self.build_group_checkboxes(region))

    def toggle_group_select(self, widget, active: bool):
        log.info(
            "log group selected",
            value=active,
            group_name=unmask_group_name(widget.get_label()),
        )
        if active:
            self.selected_groups.add(unmask_group_name(widget.get_label()))
        else:
            self.selected_groups.remove(unmask_group_name(widget.get_label()))

        create_log_group_aio_pollers(
            self.region, list(self.selected_groups), self.account
        )
        self.groups_changed(list(self.selected_groups))

    def get_state(self):
        return list(self.selected_groups)

    def build_group_checkboxes(self, region_name: str):
        return urwid.ListBox(
            [
                urwid.CheckBox(
                    abbreviate_group_name(g["name"]),
                    on_state_change=self.toggle_group_select,
                )
                for g in groups_from_region(region_name)
            ]
        )


def key_handler(sawmill):
    def show_or_exit(key):
        if key in ("q", "Q"):
            raise urwid.ExitMainLoop
        elif key in ("h", "H"):
            sawmill.help_popup()
        elif key in ("v", "V"):
            sawmill.toggle_date()
        elif key in ("c", "C"):
            sawmill.messages.clear()
        elif key in ("G"):
            # go to the last message
            sawmill.viewer.set_focus(len(sawmill.messages) - 1)
        elif key in ("g"):
            # go to the top of the page
            sawmill.viewer.set_focus(0)

    return show_or_exit


def tui_loop(region, account, backfill):
    aio_loop = asyncio.get_event_loop()
    reader = LogReadingPane(region=region, account=account, backfill=backfill)
    aio_loop.create_task(
        task_monitor(),
        name="task_monitor",
    )
    loop = urwid.MainLoop(
        reader.top,
        {
            ("timestamp", "dark red", "light gray"),
        },
        event_loop=AsyncioEventLoop(loop=aio_loop),
        unhandled_input=key_handler(reader),
    )
    loop.run()



import asyncio
import sqlite3
from datetime import datetime, timedelta, timezone
from typing import List

import boto3

from seesaw import storage
from seesaw.config import logger
from seesaw.transforms import humanize_size, localize_timestamp, relative_iso_time

log = logger(__file__)


def _pagination_coro(iterable):
    try:
        if page := next(iterable):
            return page
    except StopIteration:
        pass
    return {"events": []}


async def ordered_log_query(
    region: str,
    account_id: str,
    log_groups: List[str],
    start_from: timedelta = timedelta(minutes=5),
):
    """async coroutine that yields timestamp-ordered query from any log DB"""
    timestamp = relative_iso_time(-start_from)
    log.info(
        "Started query stack",
        region=region,
        account_id=account_id,
        log_groups=log_groups,
        start_from=start_from,
        start_ts=timestamp,
        local_start_ts=localize_timestamp(timestamp).isoformat(),
    )
    async with storage.connect(region, account_id) as db:
        events_yielded = {}
        while True:
            async for event in db.execute(
                storage.Event.select()
                .where(storage.Event.account_id == account_id)
                .where(storage.Event.timestamp > timestamp)
                .where(storage.Event.region == region)
                .where(storage.Event.log_group.in_(log_groups))
                .orderby(storage.Event.timestamp)
            ):
                if event.gid not in events_yielded:
                    yield event
                    events_yielded[event.gid] = event.timestamp

            await asyncio.sleep(1)
            if len(events_yielded) > 2000:
                expired_ts = relative_iso_time(-(start_from + timedelta(minutes=1)))
                for k in list(events_yielded.keys()):
                    if events_yielded[k] < expired_ts:
                        events_yielded.pop(k)
                log.info("Trimmed printed event cache")

            # since we poll the log groups every five seconds, give a grace
            # period of 3 polls for every paginator to get its data in
            timestamp = relative_iso_time(seconds=-15)


def groups_from_region(region_name):
    """All log group names that ARE NOT empty and older than 8 hours"""
    logs = boto3.client("logs", region_name=region_name)
    try:
        pager = (
            logs.get_paginator("describe_log_groups")
            .paginate()
            .search(
                "logGroups[].{name: logGroupName, size: storedBytes, created: creationTime}"
            )
        )
        for group in pager:
            since_created = datetime.now(tz=timezone.utc) - datetime.fromtimestamp(
                group["created"] / 1000, tz=timezone.utc
            )

            log.debug(
                "Discovered log group",
                region=region_name,
                name=group["name"],
                size=humanize_size(group.get("size") or 0),
            )

            if group["size"] or since_created < timedelta(hours=8):
                group["region"] = region_name
                yield group
    except logs.exceptions.UnrecognizedClientException:
        pass


class AsyncLogStreamPoller:
    def __init__(
        self,
        region: str,
        account_id: str,
        logs_client: boto3.client = None,
        poll_delay_seconds: int = 5,
        start_time: datetime = None,
    ):
        self.poll_delay_seconds = poll_delay_seconds
        self.region = region
        self.account_id = account_id
        self.start_time = start_time or datetime.now(tz=timezone.utc)
        self.logs = logs_client or boto3.client("logs", region_name=region)

    async def slurp_logs(
        self,
        group_name: str,
        start_time: datetime = None,
        end_time: datetime = None,
    ):
        kwargs = {
            "logGroupName": group_name,
            "interleaved": True,
        }
        if end_time is not None:
            kwargs["endTime"] = int(end_time.timestamp()) * 1000
        if start_time is None:
            kwargs["startTime"] = int(self.start_time.timestamp()) * 1000
        else:
            kwargs["startTime"] = int(start_time.timestamp()) * 1000
        while True:
            log.debug("Calling log filter", kwargs=kwargs)
            iterable = iter(
                self.logs.get_paginator("filter_log_events").paginate(**kwargs)
            )
            while True:
                page = await asyncio.get_running_loop().run_in_executor(
                    None, _pagination_coro, iterable
                )
                if not page["events"]:
                    # if a page is empty, bail
                    await asyncio.sleep(self.poll_delay_seconds)
                    break
                log.debug(
                    "Got page for",
                    group=group_name,
                    events=len(page["events"]),
                    start=kwargs["startTime"],
                    end=kwargs.get("endTime"),
                )
                incoming_events = sorted(
                    page["events"],
                    key=lambda e: e["timestamp"],
                )
                # update the start time to when we start paginating through
                kwargs["startTime"] = int(
                    max(incoming_events[-1]["timestamp"] / 1000, kwargs["startTime"])
                )
                async with storage.connect(self.region, self.account_id) as db:
                    for event in incoming_events:
                        event["logGroupName"] = group_name
                        event["region"] = self.region
                        event["accountId"] = self.account_id
                        try:
                            await db.execute(
                                storage.Event.insert().values(
                                    storage.json_to_event(event)
                                )
                            )
                        except sqlite3.IntegrityError as e:
                            if "UNIQUE constraint failed: events.gid" in repr(e):
                                # we expect to sometimes double-write events, nothing to worry about
                                continue
                            log.exception(
                                "SQLite integrity problem for event",
                                gid=event["eventId"],
                                msg=event["message"],
                                stream=event["logStreamName"],
                                group=event["logGroupName"],
                            )
                    await db.commit()
            if end_time:
                # only go through one query if there is an end time specified
                log.info(
                    "Finished backfill",
                    group=group_name,
                    start=kwargs["startTime"],
                    end=kwargs["endTime"],
                )
                break
            await asyncio.sleep(self.poll_delay_seconds)



import json
import random
from datetime import datetime, timedelta, timezone
from typing import Tuple

from dateutil import tz

localtz = tz.tzlocal()


def relative_utc_time(td: timedelta = None, **kwargs) -> datetime:
    if td and not isinstance(td, timedelta):
        raise ValueError(
            "Positional arguments are only for timedeltas, for pass-through values to timedelta us kwargs"
        )
    if td is None:
        td = timedelta(**kwargs)
    return datetime.now(tz=timezone.utc) + td


def relative_iso_time(td: timedelta = None, **kwargs) -> str:
    return relative_utc_time(td, **kwargs).isoformat()


def localize_timestamp(iso_format_dt: str) -> datetime:
    return datetime.fromisoformat(iso_format_dt).astimezone(localtz)


def abbreviate_group_name(name: str) -> str:
    if name.startswith("/aws/lambda/"):
        return name.replace("/aws/lambda/", "λ ")
    if name.startswith("/aws/vendedlogs/states/"):
        return name.replace("/aws/vendedlogs/states/", "SF ")
    if name.startswith("API-Gateway-Execution-Logs_"):
        return name.replace("API-Gateway-Execution-Logs_", "API-Exec-")
    return name


def unmask_group_name(name: str) -> str:
    if name.startswith("λ "):
        return name.replace("λ ", "/aws/lambda/")
    if name.startswith("SF "):
        return name.replace("SF ", "/aws/vendedlogs/states/")
    if name.startswith("API-Exec-"):
        return name.replace("API-Exec-", "API-Gateway-Execution-Logs_")
    return name


def random_hex():
    return "".join(random.choice("abcdef1234567890") for _ in range(10))


def humanize_size(total_size):
    for x in ("bytes", "KB", "MB", "GB", "TB"):
        if total_size < 1024.0:
            return "%.1f %s" % (total_size, x)
        total_size /= 1024.0

    return "%.1f %s" % (total_size, x)


def format_event(event, verbose_dates=True) -> Tuple[str, str]:
    ts = localize_timestamp(event.timestamp)
    if verbose_dates:
        ts = ts.strftime("%Y-%m-%d %H:%M:%S")
    else:
        ts = ts.strftime("%H:%M:%S")

    try:
        return (
            ts,
            abbreviate_group_name(event.log_group),
            json.dumps(json.loads(event.message), indent=2, sort_keys=True),
        )
    except json.decoder.JSONDecodeError:
        return ts, abbreviate_group_name(event.log_group), event.message



import os
from datetime import datetime, timedelta, timezone
from typing import Optional
from seesaw import transforms

import aql
import psutil
from aql import column
from aql.table import table as aql_table
from unsync import unsync

from seesaw.config import USER_DIR, VERSION, logger

log = logger(__file__)


def connect(region: str, account_id: str) -> aql.engines.base.Connection:
    log_db_file = "sqlite://" + str(
        USER_DIR / f"logs-{region}-{account_id}-{VERSION}.sqlite"
    )
    log.debug("Opening DB", path=log_db_file)
    mk_tables(log_db_file)
    return aql.connect(log_db_file)


@unsync
async def remove_old(region: str, account_id: str, max_age: timedelta):
    async with connect(region, account_id) as db:
        await db.execute(
            Event.delete().where(
                Event.timestamp < transforms.transforms.relative_iso_time(td=-max_age)
            )
        )


@unsync
async def mk_tables(db_file):
    connection = aql.connect(db_file)
    create_query = Event.create(if_not_exists=True)
    async with connection as db:
        await db.execute(create_query)


@aql_table("events")
class Event:
    gid: column.Primary[str]
    timestamp: str
    ingestion_time: str
    message: str
    region: str
    account_id: str
    log_group: str
    log_stream: str
    # future use, intended to help add Lambda/SF Execution ID to a special column to support selecting down to one execution
    execution_correlator: Optional[str]


def json_to_event(incoming: dict):
    return Event(
        gid=incoming["eventId"],
        account_id=incoming["accountId"],
        timestamp=datetime.fromtimestamp(
            incoming["timestamp"] / 1000, tz=timezone.utc
        ).isoformat(),
        ingestion_time=datetime.fromtimestamp(
            incoming["ingestionTime"] / 1000, tz=timezone.utc
        ).isoformat(),
        region=incoming["region"],
        message=incoming["message"],
        log_stream=incoming["logStreamName"],
        log_group=incoming["logGroupName"],
        execution_correlator=None,
    )


def get_pid_start_iso() -> str:
    """Return the iso-formatted time the current process started.

    This helps us clear out old thread signatures"""
    pid_start_ts = psutil.Process(os.getpid()).create_time()
    return (
        datetime.utcfromtimestamp(pid_start_ts).replace(tzinfo=timezone.utc).isoformat()
    )



import urwid

HELPTEXT = """
Welcome to See-Saw!

Navigation:
- Arrow keys to move left/right between sections
- Up/Down arrow and Page Up/Page Down to scroll around the logs
- Space bar to select/de-select regions and log groups

Other keybinds:
r - refresh the logs displayed based on the current log groups selected
v - turn on verbose timestamps (full date/time)
h - show this help menu
q - quit the program
""".strip()


class HelpPopUp(urwid.WidgetWrap):
    """A dialog that appears with nothing but a close button """

    signals = ["close"]

    def __init__(self):
        close_button = urwid.Button("that's pretty cool")
        urwid.connect_signal(close_button, "click", lambda button: self._emit("close"))
        pile = urwid.Pile(
            [
                urwid.Text(
                    "^^  I'm attached to the widget that opened me. "
                    "Try resizing the window!\n"
                ),
                close_button,
            ]
        )
        fill = urwid.Filler(pile)
        self.__super.__init__(urwid.AttrWrap(fill, "popbg"))
